# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import importlib.util
import json
import logging
import pathlib
from pathlib import Path
from typing import List

from vdk.internal.builtin_plugins.run.file_based_step import TYPE_PYTHON
from vdk.internal.builtin_plugins.run.job_context import JobContext
from vdk.internal.core import errors
from vdk.internal.core.errors import UserCodeError
from vdk.plugin.notebook.cell import Cell
from vdk.plugin.notebook.notebook_based_step import NotebookCellStep
from vdk.plugin.notebook.notebook_based_step import NotebookStepFuncFactory

log = logging.getLogger(__name__)


class JobNotebookLocator:
    """
    Locate the data job files that would be executed by us.
    """

    @staticmethod
    def get_notebook_files(directory: pathlib.Path) -> List[pathlib.Path]:
        """Locates the files in a directory, that are supported for execution.
        Files supported for execution are: .ipynb
        Other files in the directory are ignored.
        :return: List of files from the directory that supported for execution, sorted alphabetically by name.s
        :rtype: :class:`.list`
        """
        script_files = [
            x for x in directory.iterdir() if (x.name.lower().endswith(".ipynb"))
        ]
        script_files.sort(key=lambda x: x.name)
        log.debug(f"Script files of {directory} are {script_files}")
        return script_files


class Notebook:
    """
    Given a notebook file locates the cells with "vdk" tag and saves them.
    Files supported for reading are: ipynb.
    Other cells are ignored.
    """

    @staticmethod
    def register_notebook_steps(file_path: Path, context: JobContext):
        try:
            # see https://docs.python.org/3/library/importlib.html#importlib.util.module_from_spec
            spec = importlib.util.spec_from_loader("notebook", loader=None)
            python_module = importlib.util.module_from_spec(spec)
            # Used to declare the job_input in the new module
            # Gives access to it: module.job_inut
            # Used to pass the real vdk job_input variable in run_python_step (module.job_input = job_input
            exec("job_input = 1", python_module.__dict__)
            notebook_steps = []
            # see Jupyter json schema here:
            # https://github.com/jupyter/nbformat/blob/main/nbformat/v4/nbformat.v4.schema.json
            content = json.loads(file_path.read_text())
            index = 0
            for jupyter_cell in content["cells"]:
                if jupyter_cell["cell_type"] == "code":
                    cell = Cell(jupyter_cell)
                    if "vdk" in cell.tags:
                        index += 1
                        step = NotebookCellStep(
                            name="".join(
                                [
                                    file_path.name.replace(".ipynb", "_"),
                                    str(index),
                                ]
                            ),
                            type=cell.source_type,
                            runner_func=NotebookStepFuncFactory.get_run_function(
                                cell.source_type
                            ),
                            file_path=file_path,
                            job_dir=context.job_directory,
                            source=cell.source,
                            cell_id=cell.id,
                            module=python_module,
                        )
                        notebook_steps.append(step)
                        context.step_builder.add_step(step)
            context.step_builder._StepBuilder__steps.sort(key=lambda step: step.name)
            log.debug(f"{len(notebook_steps)} " f"cells with vdk tag were detected!")
        except json.JSONDecodeError as e:
            errors.report_and_throw(
                UserCodeError(
                    f"Failed to read the {file_path.name} file.",
                    f"The provided {file_path.name} cannot be loaded into json format and "
                    f"cannot be read as a Jupyter notebook",
                    errors.MSG_CONSEQUENCE_TERMINATING_APP,
                    f"Check the {file_path.name} format again",
                )
            )



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import os
import unittest
from unittest import mock

from click.testing import Result
from vdk.plugin.notebook import notebook_plugin
from vdk.plugin.sqlite import sqlite_plugin
from vdk.plugin.test_utils.util_funcs import cli_assert_equal
from vdk.plugin.test_utils.util_funcs import CliEntryBasedTestRunner
from vdk.plugin.test_utils.util_funcs import jobs_path_from_caller_directory


@mock.patch.dict(
    os.environ,
    {
        "VDK_DB_DEFAULT_TYPE": "SQLITE",
        "VDK_INGEST_METHOD_DEFAULT": "sqlite",
    },
)
class JupyterTests(unittest.TestCase):
    def setUp(self) -> None:
        self.__runner = CliEntryBasedTestRunner(notebook_plugin, sqlite_plugin)

    def test_successful_job(self) -> None:
        result: Result = self.__runner.invoke(
            ["run", jobs_path_from_caller_directory("ingest-job")]
        )
        cli_assert_equal(0, result)
        actual_rs: Result = self.__runner.invoke(
            ["sqlite-query", "--query", "SELECT * FROM rest_target_table"]
        )
        assert actual_rs.stdout == (
            "  userId    id  title                 completed\n"
            "--------  ----  ------------------  -----------\n"
            "       1     1  delectus aut autem            0\n"
        )

    def test_successful_job_magic_cells(self) -> None:
        result: Result = self.__runner.invoke(
            ["run", jobs_path_from_caller_directory("ingest-job-sql-magic-cells")]
        )
        cli_assert_equal(0, result)
        actual_rs: Result = self.__runner.invoke(
            ["sqlite-query", "--query", "SELECT * FROM rest_target_table_magic"]
        )
        assert actual_rs.stdout == (
            "  userId    id  title                 completed\n"
            "--------  ----  ------------------  -----------\n"
            "       1     1  delectus aut autem            0\n"
        )

    def test_successful_job_magic_cells_broken(self) -> None:
        result: Result = self.__runner.invoke(
            [
                "run",
                jobs_path_from_caller_directory("ingest-job-sql-magic-cells-broken"),
            ]
        )
        cli_assert_equal(1, result)
        assert "SyntaxError" in result.output

    def test_failing_job_with_syntax_error(self) -> None:
        result: Result = self.__runner.invoke(
            ["run", jobs_path_from_caller_directory("ingest-job-fail-syntax-error")]
        )
        cli_assert_equal(1, result)

    def test_failing_job_with_code_error(self) -> None:
        result: Result = self.__runner.invoke(
            ["run", jobs_path_from_caller_directory("ingest-job-fail-code-error")]
        )
        cli_assert_equal(2, result)

    def test_failing_job_with_sql_error(self) -> None:
        result: Result = self.__runner.invoke(
            ["run", jobs_path_from_caller_directory("ingest-job-sql-error")]
        )
        cli_assert_equal(1, result)

    def test_mixed_job_with_py_and_sql(self) -> None:
        result: Result = self.__runner.invoke(
            ["run", jobs_path_from_caller_directory("mixed-job")]
        )
        cli_assert_equal(0, result)
        actual_rs: Result = self.__runner.invoke(
            ["sqlite-query", "--query", "SELECT * FROM rest_target_table"]
        )
        assert actual_rs.stdout == (
            "  userId    id  title                 completed\n"
            "--------  ----  ------------------  -----------\n"
            "       1     1  delectus aut autem            0\n"
        )



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import json
import logging

from vdk.api.plugin.hook_markers import hookimpl
from vdk.api.plugin.plugin_registry import HookCallResult
from vdk.api.plugin.plugin_registry import IPluginRegistry
from vdk.internal.builtin_plugins.run.execution_results import ExecutionResult
from vdk.internal.builtin_plugins.run.job_context import JobContext
from vdk.internal.builtin_plugins.run.run_status import ExecutionStatus
from vdk.plugin.notebook.notebook import JobNotebookLocator
from vdk.plugin.notebook.notebook import Notebook

log = logging.getLogger(__name__)


class NotebookPlugin:
    @hookimpl(trylast=True)
    def initialize_job(self, context: JobContext):
        file_locator: JobNotebookLocator = JobNotebookLocator()
        notebook_files = file_locator.get_notebook_files(context.job_directory)
        if len(notebook_files) >= 1:
            for file_path in notebook_files:
                Notebook.register_notebook_steps(file_path, context)


@hookimpl
def vdk_start(plugin_registry: IPluginRegistry):
    plugin_registry.load_plugin_with_hooks_impl(NotebookPlugin(), "notebook-plugin")



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from dataclasses import dataclass

from vdk.internal.builtin_plugins.run.file_based_step import TYPE_PYTHON
from vdk.internal.builtin_plugins.run.file_based_step import TYPE_SQL


@dataclass
class Cell:
    """
    A class that represents Jupyter code cell
    See Jupyter cell in this json schema:
             https://github.com/jupyter/nbformat/blob/main/nbformat/v4/nbformat.v4.schema.json
    """

    def __init__(self, jupyter_cell):
        self.tags = jupyter_cell["metadata"].get("tags", {})
        self.source, self.source_type = self.__extract_source_code(jupyter_cell)
        self.id = jupyter_cell.get("id")

    @staticmethod
    def __extract_source_code(jupyter_cell):
        lines = jupyter_cell["source"]
        if lines and lines[0].strip().startswith("%%vdksql"):
            statement = "".join(lines[1:])
            return statement, TYPE_SQL
        return "".join(lines), TYPE_PYTHON



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import pathlib

import setuptools


__version__ = "0.1.1049090626"

setuptools.setup(
    name="vdk-notebook",
    version=__version__,
    url="https://github.com/vmware/versatile-data-kit",
    description="A VDK plugin for working with notebooks",
    long_description=pathlib.Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    install_requires=["vdk-core"],
    package_dir={"": "src"},
    packages=setuptools.find_namespace_packages(where="src"),
    entry_points={"vdk.plugin.run": ["notebook = vdk.plugin.notebook.notebook_plugin"]},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    python_requires="~=3.7",
)



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
import sys
import traceback
from dataclasses import dataclass
from typing import Callable
from typing import Union

from vdk.api.job_input import IJobInput
from vdk.internal.builtin_plugins.run.file_based_step import TYPE_PYTHON
from vdk.internal.builtin_plugins.run.file_based_step import TYPE_SQL
from vdk.internal.builtin_plugins.run.step import Step
from vdk.internal.core import errors
from vdk.internal.core.errors import UserCodeError

log = logging.getLogger(__name__)

# consists may duplicates of
# https://github.com/vmware/versatile-data-kit/blob/main/projects/vdk-core/src/vdk/internal/builtin_plugins/run/file_based_step.py

# The function accept NotebookCellStep (below class) and IJobInput and
# return true if the step has been executed and false if it is not (valid) executable step.
# On error it's expected to raise an exception.
NotebookStepFunction = Callable[["NotebookCellStep", IJobInput], bool]


@dataclass
class NotebookCellStep(Step):
    """
    A notebook step that will be executed when running a data job.

    Step class attributes:
    ::name: str - the name of the concrete step (e.g name of the file)
    ::type: str - string representing a step type (sql or python).
    ::runner_func: StepFunction - function that will execute the actual step
    ::file_path: pathlib.Path - file where the step is defined
    ::job_dir: pathlib.Path - the root job directory
    ::parent: Step | None = None - parent Step

    Additional attributes:
    ::source: str - the code string retrieved from Jupyter code cell
    ::module: module object - the module the code belongs to
    (see imp.new_module in https://docs.python.org/3/library/imp.html)
    """

    def __init__(
        self,
        name,
        type,
        runner_func,
        file_path,
        job_dir,
        source,
        cell_id,
        module=None,
        parent=None,
    ):
        super().__init__(name, type, runner_func, file_path, job_dir, parent)
        self.runner_func = runner_func
        self.source = source
        self.module = module
        self.cell_id = cell_id


class NotebookStepFuncFactory:
    """
    Implementations of runner_func for running Notebook steps
    """

    @staticmethod
    def get_run_function(source_type: Union[TYPE_PYTHON, TYPE_SQL]) -> Callable:
        if source_type == TYPE_PYTHON:
            return NotebookStepFuncFactory.run_python_step
        elif source_type == TYPE_SQL:
            return NotebookStepFuncFactory.run_sql_step
        else:
            raise NotImplementedError(
                f"Run function for source type {source_type} is not implemented."
            )

    @staticmethod
    def run_sql_step(step: NotebookCellStep, job_input: IJobInput) -> bool:
        """ """
        job_input.execute_query(step.source)
        return True

    @staticmethod
    def run_python_step(step: NotebookCellStep, job_input: IJobInput) -> bool:
        try:
            sys.path.insert(0, str(step.job_dir))
            success = False
            try:
                log.debug("Loading %s ..." % step.name)
                step.module.job_input = job_input
                exec(step.source, step.module.__dict__)
                log.debug("Loading %s SUCCESS" % step.name)
                success = True
            except SyntaxError as e:
                log.info("Loading %s FAILURE" % step.name)
                errors.report_and_throw(
                    UserCodeError(
                        f"Failed loading job sources of {step.name} from cell with cell_id:{step.cell_id}"
                        f" from {step.file_path.name}",
                        f"{e.__class__.__name__} at line {e.lineno} of {step.name}"
                        f": {e.args[0]}",
                        f"Current Step {step.name} from {step.file_path}"
                        "will fail, and as a result the whole Data Job will fail. ",
                        f"Please, check the {step.file_path.name} file again for syntax errors",
                        f"{e}",
                    )
                )
            except Exception as e:
                cl, exc, tb = sys.exc_info()
                line_number = traceback.extract_tb(tb)[-1][1]
                errors.report_and_throw(
                    UserCodeError(
                        f"Failed loading job sources of {step.name} from cell with cell_id:{step.cell_id}"
                        f" from {step.file_path.name}",
                        f"{e.__class__.__name__} at line {line_number} of {step.name}"
                        f": {e.args[0]}",
                        f"Current Step {step.name} from {step.file_path}"
                        "will fail, and as a result the whole Data Job will fail. ",
                        f"Please, check the {step.file_path.name} file again for errors",
                        f"{e}",
                    )
                )
            return success
        finally:
            sys.path.remove(str(step.job_dir))
